/* 
** x86_video.c:
**
** Christophe Lucas <clucas@rotomalug.org>
**
*/

#include <video/x86_video.h>
#include <testos/ioports.h>
#include <testos/types.h>


/* Some screen stuff.  */
#define COLUMNS			80	/* The number of columns.  */
#define LINES			24	/* The number of lines.  */
#define ATTRIBUTE		7	/* The attribute of an character.  */
#define VIDEO			0xB8000	/* The video memory address.  */

/* Variables.  */
static int xpos;			/* Save the X position.  */
static int ypos;			/* Save the Y position.  */
static volatile unsigned char *video;	/* Point to the video memory.  */
/* 
 * The structure of a character element in the video memory. @see
 * http://webster.cs.ucr.edu/AoA DOS edition chapter 23 
 */
/* From SOS :
typedef struct {
	unsigned char character;
        unsigned char attribute;
} __attribute__ ((packed)) x86_video_mem[LINES*COLUMNS];
*/

ret_t x86_video_setup(void)
{
	/*
	 * Hide cursor. @see Ralf Brown's interrupt (and port) list
	 * http://www-2.cs.cmu.edu/~ralf/files.html
	 */
#define CRT_REG_INDEX 0x3d4
#define CRT_REG_DATA  0x3d5

	/* CRT index port => ask for access to register 0xa ("cursor start") */
	outb(0x0a, CRT_REG_INDEX);

	/* (RBIL Tables 708 & 654) CRT Register 0xa => bit 5 = cursor OFF */
	outb(1 << 5, CRT_REG_DATA);

	return OK;
}
			  


/* Clear the screen and initialize VIDEO, XPOS and YPOS.  */
void cls (void)
{
	int i;

	video = (unsigned char *) VIDEO;
  
	for (i = 0; i < COLUMNS * LINES * 2; i++)
		*(video + i) = 0;

	xpos = 0;
	ypos = 0;
}

/* Convert the integer D to a string and save the string in BUF. If
   BASE is equal to 'd', interpret that D is decimal, and if BASE is
   equal to 'x', interpret that D is hexadecimal.  */
static void itoa (char *buf, int base, int d)
{
	char *p = buf;
	char *p1, *p2;
	unsigned long ud = d;
	int divisor = 10;
  
	/* If %d is specified and D is minus, put `-' in the head.  */
	if (base == 'd' && d < 0) {
		*p++ = '-';
		buf++;
		ud = -d;
	} else if (base == 'x')
		divisor = 16;

	/* Divide UD by DIVISOR until UD == 0.  */
	do {
		int remainder = ud % divisor;
      		*p++ = (remainder < 10) ? remainder + '0' : remainder + 'a' - 10;
	} while (ud /= divisor);

	/* Terminate BUF.  */
	*p = 0;
  
	/* Reverse BUF.  */
	p1 = buf;
	p2 = p - 1;
	while (p1 < p2) {
		char tmp = *p1;
		*p1 = *p2;
		*p2 = tmp;
		p1++;
		p2--;
	}
}

/* Put the character C on the screen.  */
static void putchar (int c)
{
	if (c == '\n' || c == '\r') {
newline:
		xpos = 0;
		ypos++;
		if (ypos >= LINES)
			ypos = 0;
		return;
	}

	*(video + (xpos + ypos * COLUMNS) * 2) = c & 0xFF;
	*(video + (xpos + ypos * COLUMNS) * 2 + 1) = ATTRIBUTE;

	xpos++;
	if (xpos >= COLUMNS)
		goto newline;
}

/*
 * Format a string and print it on the screen, just like the libc
 * function printf.  
 */
void printk (const char *format, ...)
{
	char **arg = (char **) &format;
	int c;
	char buf[20];

	arg++;
  
	while ((c = *format++) != 0) {
		if (c != '%')
			putchar (c);
		else {
			char *p;
	  
			c = *format++;
			switch (c) {
	    			case 'd':
				case 'u':
				case 'x':
					itoa (buf, c, *((int *) arg++));
					p = buf;
					goto string;
				break;

				case 's':
					p = *arg++;
					if (! p)
						p = "(null)";
			    string:
					while (*p)
						putchar (*p++);
				break;

				default:
					putchar (*((int *) arg++));
				break;
			} /* switch */
		} /* else */
	} /* while */
}
